@@ -50,4 +50,17 @@ std::string binaryToString(uint64_t x, size_t len) {
     s.at(i) = c;
   }
   return std::move(s);
-}
\ No newline at end of file
+}
+
+int hamming(uint64_t a, uint64_t b, size_t len) {
+  uint64_t df = a^b;
+  int d = 0;
+  size_t sh = len-1;
+  for (size_t i = 0; i < len; ++i) {
+    if (((df>>(2*sh)) & 0x03ULL) != 0) {
+      d++;
+    }    
+    sh--;
+  }
+  return d;
+}
