@@ -1,7 +1,7 @@
 #ifndef KALLISTO_COMMON_H
 #define KALLISTO_COMMON_H
 
-#define KALLISTO_VERSION "0.46.2"
+#define KALLISTO_VERSION "0.48.0"
 
 #include <string>
 #include <vector>
@@ -22,9 +22,11 @@ struct BUSOptionSubstr {
 struct BUSOptions {
   int nfiles;
   
-  BUSOptionSubstr umi;
+  std::vector<BUSOptionSubstr> umi;
   std::vector<BUSOptionSubstr> bc;
   std::vector<BUSOptionSubstr> seq;
+  
+  bool paired;
 
   int getBCLength() const {
     int r =0 ;
@@ -32,6 +34,8 @@ struct BUSOptions {
       for (auto& b : bc) {
         if (b.start < 0) {
           return 0;
+        } else if (b.stop == 0) {
+          return 0;
         } else {
           r += b.stop - b.start;
         }
@@ -41,11 +45,19 @@ struct BUSOptions {
   }
 
   int getUMILength() const {
-    if (umi.start >= 0) {
-      return umi.stop - umi.start;
-    } else {
-      return 0;
+    int r =0 ;
+    if (!umi.empty()) {
+      for (auto& u : umi) {
+        if (u.start < 0) {
+          return 0;
+        } else if (u.stop == 0) {
+          return 0;
+        } else {
+          r += u.stop - u.start;
+        }
+      }
     }
+    return r;
   }
 };
 
@@ -64,6 +76,7 @@ struct ProgramOptions {
   int bootstrap;
   std::vector<std::string> transfasta;
   bool batch_mode;
+  bool pseudo_read_files_supplied;
   bool bus_mode;
   BUSOptions busOptions;
   bool pseudo_quant;
@@ -87,6 +100,8 @@ struct ProgramOptions {
   enum class StrandType {None, FR, RF};
   StrandType strand;
   bool umi;
+  bool batch_bus_write;
+  bool batch_bus;
   std::string gfa; // used for inspect
   bool inspect_thorough;
   bool single_overhang;
@@ -94,6 +109,12 @@ struct ProgramOptions {
   std::string chromFile;
   std::string bedFile;
   std::string technology;
+  std::string tagsequence;
+  std::string tccFile;
+  std::string ecFile;
+  std::string fldFile;
+  std::string transcriptsFile;
+  std::string genemap;
 
 ProgramOptions() :
   verbose(false),
@@ -107,6 +128,7 @@ ProgramOptions() :
   min_range(1),
   bootstrap(0),
   batch_mode(false),
+  pseudo_read_files_supplied(false),
   bus_mode(false),
   pseudo_quant(false),
   bam(false),
@@ -123,6 +145,8 @@ ProgramOptions() :
   fusion(false),
   strand(StrandType::None),
   umi(false),
+  batch_bus_write(false),
+  batch_bus(false),
   inspect_thorough(false),
   single_overhang(false)
   {}
